import re

translation_table = {
    "UUU": "Phe",
    "UUC": "Phe",
    "UUA": "Leu",
    "UUG": "Leu",
    "CUU": "Leu",
    "CUC": "Leu",
    "CUA": "Leu",
    "CUG": "Leu",
    "AUU": "Ile",
    "AUC": "Ile",
    "AUA": "Ile",
    "AUG": "Met",
    "GUU": "Val",
    "GUC": "Val",
    "GUA": "Val",
    "GUG": "Val",
    "UCU": "Ser",
    "UCC": "Ser",
    "UCA": "Ser",
    "UCG": "Ser",
    "CCU": "Pro",
    "CCC": "Pro",
    "CCA": "Pro",
    "CCG": "Pro",
    "ACU": "Thr",
    "ACC": "Thr",
    "ACA": "Thr",
    "ACG": "Thr",
    "GCU": "Ala",
    "GCC": "Ala",
    "GCA": "Ala",
    "GCG": "Ala",
    "UAU": "Tyr",
    "UAC": "Tyr",
    "UAA": "",
    "UAG": "",
    "CAU": "His",
    "CAC": "His",
    "CAA": "Gln",
    "CAG": "Gln",
    "AAU": "Asn",
    "AAC": "Asn",
    "AAA": "Lys",
    "AAG": "Lys",
    "GAU": "Asp",
    "GAC": "Asp",
    "GAA": "Glu",
    "GAG": "Glu",
    "UGU": "Cys",
    "UGC": "Cys",
    "UGA": "",
    "UGG": "Trp",
    "CGU": "Arg",
    "CGC": "Arg",
    "CGA": "Arg",
    "CGG": "Arg",
    "AGU": "Ser",
    "AGC": "Ser",
    "AGA": "Arg",
    "AGG": "Arg",
    "GGU": "Gly",
    "GGC": "Gly",
    "GGA": "Gly",
    "GGG": "Gly",
}


class DNA_obj:
    def __init__(self, seq: str, direction: str = "1"):
        """
        Instantiates a DNA_obj class object.

        :param seq: A string that holds a DNA sequence.
        :type seq: str
        :param direction: A string that holds the DNA sequence's orientation. "1" for 5'->3' and "2" for 3'->5' (default is "1").
        :type direction: str
        """
        self.seq: str = seq.strip().upper()
        self.direction: str = direction.strip()
        self.complementary: str = ""
        self.rna_seq: str = ""
        self.peptide_seq: str = ""

    def __str__(self):
        """
        Returns a string representation of a DNA_obj class instance.
        The representation contains the following:
        1. The DNA's coding strand sequence.
        2. The complementary DNA's strand sequence.
        3. The mRNA sequence.
        4. The amino acid sequence.

        :return: A string that represents a DNA_obj object.
        :rtype: str  
        """
        # \u2082 is utf 8 encoding specification for 'subscript 2' and works in a f string
        return f"DNA Coding Strand:         5' {self.justify_nucleotide_seq(self.seq)} 3'\nDNA Complementary Strand:  3' {self.justify_nucleotide_seq(self.complementary)} 5'\nmRNA sequence:             5' {self.justify_nucleotide_seq(self.rna_seq)} 3'\nPeptide sequence:         H\u2082N-{self.peptide_seq}-COOH"

    # seq getter
    @property
    def seq(self):
        """
        Gets the DNA sequence from the corresponding DNA_obj class parameter.

        :return: A string that holds the DNA object's sequence.
        :rtype: str
        """
        return self._seq

    # seq setter
    @seq.setter
    def seq(self, seq: str):
        """
        Sets the DNA object's sequence. This function runs automatically when the class _init_ function is called.

        :param seq: A string that holds the DNA sequence.
        :type seq: str
        :raises ValueError: If an invalid letter exists in the DNA sequence.
        """
        if re.search(r"[^atgcATGC]", seq):
            raise ValueError("Invalid Sequence!")
        else:
            self._seq = seq

    # direction getter
    @property
    def direction(self):
        """
        Gets the DNA sequence's orientation from the corresponding DNA_obj class parameter.
        :return: A string that holds the DNA sequence's orientation. "1" for 5'->3' and "2" for 3'->5'.
        :rtype: str
        """
        return self._direction

    # direction setter
    @direction.setter
    def direction(self, direction: str):
        """
        Sets the DNA sequence's orientation. This function runs automatically when the class _init_ function is called.

        :param direction: A string that holds the DNA sequence orientation. "1" for 5'->3' and "2" for 3'->5'.
        :type direction: str
        :raise ValueError: If the orientation is not set as "1" or "2"
        """
        if direction != "1" and direction != "2":
            raise ValueError("Selection is not valid.")
        else:
            self._direction = direction

    def set_seq_dir(self):
        """
        Sets the orientation of a DNA sequence
        For orientation "1", which corresponds to 5'->3', the function keeps the given DNA sequence string as is
        For orientation "2", which corresponds to 3'->5', the function reverses the DNA sequence string

        :raise ValueError: If inputed orientation is not "1" or "2"
        """
        if self.direction == "2":
            self.seq = self.seq[::-1]
        elif self.direction == "1":
            self.seq = self.seq
        else:
            raise ValueError("Selection is not valid.")

    def comp(self):
        """
        Creates the complementary DNA sequence for a given DNA sequence
        """
        self.complementary = self.seq.replace("T", "a")
        self.complementary = self.complementary.replace("A", "t")
        self.complementary = self.complementary.replace("C", "g")
        self.complementary = self.complementary.replace("G", "c").upper()

    def transcription(self):
        """
        Creates the corresponding mRNA sequence for a given DNA sequence
        """
        self.rna_seq = self.seq.replace("T", "U")

    def translation(self):
        """
        Translates a DNA nucleotide sequence to a "-" delimited amino acid sequence
        """
        if re.search(r"^ATG(?:[ATGC]{3})*(?:TGA|TAA|TAG)$", self.seq):
            i = 0
            while i < len(self.rna_seq) - 2:
                if len(translation_table[self.rna_seq[i : i + 3]]) == 0:
                    break
                self.peptide_seq += f"{translation_table[self.rna_seq[i:i+3]]}-"
                i += 3
            self.peptide_seq = self.peptide_seq[:-1]
        else:
            self.peptide_seq = "Non Coding"

    def justify_nucleotide_seq(self, seq: str):
        """
        Adds a "-" delimiter every 3 letters of a nucleotide sequence and return the justified (delimited) sequence

        :param seq: A string that holds the nucleotide sequence
        :type seq: str
        :return: A string that holds the justified sequence
        :rtype: str
        """
        i = 0
        just_nuc_seq = ""
        while i < len(seq):
            just_nuc_seq += f"{seq[i:i+3]}-"
            i += 3
        # return all elements of the justified nucleotide sequence, except the last -
        return just_nuc_seq[:-1]
